-- Specify that everything is happening in the UBS DB. Must be in every SQL init script!
\c ubs;

CREATE EXTENSION IF NOT EXISTS pgcrypto;

-- FILE

CREATE TABLE IF NOT EXISTS file_index (
  id BIGSERIAL PRIMARY KEY NOT NULL,
  sha256 bytea NOT NULL CHECK(length(sha256) = 32),
  md5 bytea NOT NULL CHECK(length(md5) = 16),
  size BIGINT NOT NULL,
  date_created TIMESTAMPTZ NOT NULL DEFAULT NOW()
);

CREATE UNIQUE INDEX IF NOT EXISTS idx_file_sha256 ON file_index (sha256);
CREATE INDEX IF NOT EXISTS idx_file_md5 ON file_index (md5);
CREATE INDEX IF NOT EXISTS idx_file_index_date_created ON file_index (date_created);

CREATE TABLE IF NOT EXISTS file_details (
  id BIGSERIAL PRIMARY KEY NOT NULL,
  file_index_id BIGINT NOT NULL REFERENCES file_index(id),
  sha256 bytea NOT NULL CHECK(length(sha256) = 32),
  md5 bytea NOT NULL CHECK(length(md5) = 16),
  s3_key varchar(1024) NOT NULL,
  size BIGINT NOT NULL,
  truncated BOOLEAN NOT NULL DEFAULT FALSE,
  date_created TIMESTAMPTZ NOT NULL DEFAULT NOW(),
  date_modified TIMESTAMPTZ NOT NULL DEFAULT NOW(),
  test BOOLEAN NOT NULL DEFAULT FALSE
);

CREATE UNIQUE INDEX IF NOT EXISTS idx_file_details_file_index_id ON file_details (file_index_id);

CREATE TABLE IF NOT EXISTS program_name (
  id BIGSERIAL PRIMARY KEY NOT NULL,
  name VARCHAR,
  sha256_name bytea NOT NULL
);

CREATE UNIQUE INDEX IF NOT EXISTS idx_program_name_sha256_name ON program_name(sha256_name);

CREATE TABLE IF NOT EXISTS company_name (
  id BIGSERIAL PRIMARY KEY NOT NULL,
  name VARCHAR,
  sha256_name bytea NOT NULL
);

CREATE UNIQUE INDEX IF NOT EXISTS idx_company_name_sha256_name ON company_name(sha256_name);

CREATE TABLE IF NOT EXISTS copyright (
  id BIGSERIAL PRIMARY KEY NOT NULL,
  name VARCHAR,
  sha256_name bytea NOT NULL
);

CREATE UNIQUE INDEX IF NOT EXISTS idx_copyright_sha256_name ON copyright(sha256_name);

CREATE TABLE IF NOT EXISTS metadata (
  id BIGSERIAL PRIMARY KEY NOT NULL,
  file_index_id BIGINT NOT NULL REFERENCES file_index(id),
  internal_name_id BIGINT NOT NULL REFERENCES program_name(id),
  product_name_id BIGINT NOT NULL REFERENCES program_name(id),
  company_name_id BIGINT NOT NULL REFERENCES company_name(id),
  copyright_id BIGINT NOT NULL REFERENCES copyright(id),
  trademark_id BIGINT NOT NULL REFERENCES copyright(id),
  file_description VARCHAR,
  file_version VARCHAR,
  comments VARCHAR,
  original_filename VARCHAR,
  product_description VARCHAR,
  product_version VARCHAR,
  private_build VARCHAR,
  special_build VARCHAR,
  icon bytea,
  os_type VARCHAR,
  architecture SMALLINT,
  lang_id INT,
  charset_id INT,
  sensor_version VARCHAR,
  internal_version SMALLINT NOT NULL DEFAULT 1,
  policy_version SMALLINT NOT NULL DEFAULT 1
);

CREATE UNIQUE INDEX IF NOT EXISTS idx_metadata_file_index_id ON metadata(file_index_id);
CREATE INDEX IF NOT EXISTS idx_metadata_internal_name_id ON metadata(internal_name_id);
CREATE INDEX IF NOT EXISTS idx_metadata_product_name_id ON metadata(product_name_id);
CREATE INDEX IF NOT EXISTS idx_metadata_company_name_id ON metadata(company_name_id);
CREATE INDEX IF NOT EXISTS idx_metadata_copyright_id ON metadata(copyright_id);
CREATE INDEX IF NOT EXISTS idx_metadata_trademark_id ON metadata(trademark_id);

CREATE TABLE IF NOT EXISTS signature (
  id BIGSERIAL PRIMARY KEY NOT NULL,
  digsig_publisher_name_id BIGINT NOT NULL REFERENCES company_name(id),
  digsig_issuer_name_id BIGINT NOT NULL REFERENCES company_name(id),
  digsig_sign_time TIMESTAMPTZ,
  date_created TIMESTAMPTZ NOT NULL DEFAULT NOW()
);

CREATE UNIQUE INDEX IF NOT EXISTS idx_signature_publisher_name_issuer_name_sign_time_not_null
  ON signature(digsig_publisher_name_id, digsig_issuer_name_id, digsig_sign_time) WHERE digsig_sign_time IS NOT NULL;
CREATE UNIQUE INDEX IF NOT EXISTS idx_signature_publisher_name_issuer_name_null_sign_time
  ON signature(digsig_publisher_name_id, digsig_issuer_name_id) WHERE digsig_sign_time IS NULL;
CREATE INDEX IF NOT EXISTS idx_signature_digsig_publisher_name_id ON signature(digsig_publisher_name_id);
CREATE INDEX IF NOT EXISTS idx_signature_digsig_issuer_name_id ON signature(digsig_issuer_name_id);
CREATE INDEX IF NOT EXISTS idx_signature_digsig_sign_time ON signature(digsig_sign_time);

CREATE TABLE IF NOT EXISTS partner (
  id SMALLSERIAL PRIMARY KEY NOT NULL,
  name VARCHAR(128) NOT NULL,
  date_created TIMESTAMPTZ NOT NULL DEFAULT NOW()
);

INSERT INTO partner (id, name) VALUES (0, 'apc') ON CONFLICT DO NOTHING;

CREATE UNIQUE INDEX IF NOT EXISTS idx_partner_name ON partner(name);

CREATE TABLE IF NOT EXISTS file_details_partner_link (
  id BIGSERIAL PRIMARY KEY NOT NULL,
  partner_id SMALLINT NOT NULL REFERENCES partner(id),
  file_details_id BIGINT NOT NULL REFERENCES file_details(id),
  response_state SMALLINT NOT NULL,
  date_created TIMESTAMPTZ NOT NULL DEFAULT NOW(),
  date_modified TIMESTAMPTZ NOT NULL DEFAULT NOW()
);

CREATE INDEX IF NOT EXISTS idx_file_details_partner_link_partner_id ON file_details_partner_link(partner_id);
CREATE INDEX IF NOT EXISTS idx_file_details_partner_link_file_details_id ON file_details_partner_link(file_details_id);

CREATE TABLE IF NOT EXISTS magic (
  id SERIAL PRIMARY KEY NOT NULL,
  value varchar(1024) NOT NULL,
  sha256_value bytea NOT NULL CHECK(length(sha256_value) = 32) UNIQUE,
  date_created TIMESTAMPTZ NOT NULL DEFAULT NOW()
);

CREATE INDEX IF NOT EXISTS idx_magic_sha256_value ON magic (sha256_value);

CREATE TABLE IF NOT EXISTS file_details_magic_link (
  id BIGSERIAL PRIMARY KEY NOT NULL,
  magic_id INT NOT NULL REFERENCES magic(id),
  file_details_id BIGINT NOT NULL REFERENCES file_details(id),
  date_created TIMESTAMPTZ NOT NULL DEFAULT NOW()
);

CREATE INDEX IF NOT EXISTS idx_file_details_magic_link_magic_id ON file_details_magic_link (magic_id);
CREATE UNIQUE INDEX IF NOT EXISTS idx_file_details_magic_link_file_index_id ON file_details_magic_link (file_details_id);

-- CUST

CREATE TABLE IF NOT EXISTS cust (
  id BIGSERIAL UNIQUE NOT NULL PRIMARY KEY,
  search_key bytea NOT NULL CHECK(length(search_key) = 32), --should be digest(org_key, 'sha256')
  date_created TIMESTAMPTZ NOT NULL DEFAULT NOW(),
  org_key VARCHAR(128)
);

CREATE UNIQUE INDEX IF NOT EXISTS idx_cust_search_key ON cust(search_key);

CREATE TABLE IF NOT EXISTS cust_ix_link (
  id BIGSERIAL PRIMARY KEY NOT NULL,
  file_index_id BIGINT NOT NULL REFERENCES file_index(id),
  cust_id BIGINT NOT NULL REFERENCES cust(id),
  date_created TIMESTAMPTZ NOT NULL DEFAULT NOW()
);

-- This pair should be unique but risk is low since we only select top, and performance is prioritized.
CREATE UNIQUE INDEX IF NOT EXISTS idx_cust_ix_link_file_index_id_cust_id ON cust_ix_link (file_index_id, cust_id);
CREATE INDEX IF NOT EXISTS idx_cust_ix_link_cust_id ON cust_ix_link (cust_id);
CREATE INDEX IF NOT EXISTS idx_cust_ix_link_date_created ON cust_ix_link (date_created);

CREATE TABLE IF NOT EXISTS file_details_cust_link (
  id BIGSERIAL PRIMARY KEY NOT NULL,
  file_details_id BIGINT NOT NULL REFERENCES file_details(id),
  cust_id BIGINT NOT NULL REFERENCES cust(id),
  date_created TIMESTAMPTZ NOT NULL DEFAULT NOW()
);

CREATE INDEX IF NOT EXISTS idx_file_details_cust_link_file_details_id ON file_details_cust_link (file_details_id);
CREATE INDEX IF NOT EXISTS idx_file_details_cust_link_cust_id ON file_details_cust_link (cust_id);

CREATE TABLE IF NOT EXISTS file_index_cust_partner_link (
  id BIGSERIAL PRIMARY KEY NOT NULL,
  file_index_id BIGINT NOT NULL REFERENCES file_index(id),
  cust_id BIGINT NOT NULL REFERENCES cust(id),
  partner_id SMALLINT NOT NULL REFERENCES partner(id)
);

CREATE UNIQUE INDEX IF NOT EXISTS idx_file_index_cust_partner_link_file_index_id_partner_id
  ON file_index_cust_partner_link (file_index_id, partner_id);

CREATE TABLE IF NOT EXISTS signature_result_link (
  id BIGSERIAL PRIMARY KEY NOT NULL,
  cust_ix_link_id BIGINT NOT NULL REFERENCES cust_ix_link(id),
  signature_id BIGINT NOT NULL REFERENCES signature(id),
  is_catalog_signature BOOLEAN NOT NULL DEFAULT FALSE,
  first_seen TIMESTAMPTZ NOT NULL,
  num_occurrences INT NOT NULL DEFAULT 1,
  date_created TIMESTAMPTZ NOT NULL DEFAULT NOW(),
  date_modified TIMESTAMPTZ NOT NULL DEFAULT NOW(),
  sig_state SMALLINT NOT NULL DEFAULT 0
);

CREATE UNIQUE INDEX IF NOT EXISTS idx_signature_result_link_cust_ix_link_id_signature_id ON signature_result_link(cust_ix_link_id, signature_id, sig_state);
CREATE INDEX IF NOT EXISTS idx_signature_result_link_is_catalog_signature ON signature_result_link(is_catalog_signature);

CREATE TABLE IF NOT EXISTS device (
  id BIGSERIAL PRIMARY KEY NOT NULL,
  device_id BIGINT NOT NULL,
  device_name VARCHAR(256) NOT NULL,
  date_created TIMESTAMPTZ NOT NULL DEFAULT NOW()
);

CREATE UNIQUE INDEX IF NOT EXISTS idx_device_device_id_device_name ON device(device_id, device_name);

ALTER TABLE device ADD COLUMN IF NOT EXISTS last_seen TIMESTAMPTZ DEFAULT NULL;

CREATE TABLE IF NOT EXISTS cust_device_link (
  id BIGSERIAL PRIMARY KEY NOT NULL,
  cust_ix_link_id BIGINT NOT NULL REFERENCES cust_ix_link(id),
  device_id BIGINT NOT NULL REFERENCES device(id),
  first_seen TIMESTAMPTZ NOT NULL
);

CREATE UNIQUE INDEX IF NOT EXISTS idx_cust_device_link_cust_ix_link_id_device_id ON cust_device_link(cust_ix_link_id, device_id);

-- cust_device_link_1 ---
CREATE TABLE IF NOT EXISTS cust_device_link_1 (
  id BIGSERIAL PRIMARY KEY NOT NULL,
  cust_ix_link_id BIGINT NOT NULL REFERENCES cust_ix_link(id),
  device_id BIGINT NOT NULL REFERENCES device(id),
  first_seen TIMESTAMPTZ NOT NULL
);

CREATE UNIQUE INDEX IF NOT EXISTS idx_cust_device_link_1_cust_ix_link_id_device_id ON cust_device_link_1(cust_ix_link_id, device_id);

-- cust_device_link_3 ---

CREATE TABLE IF NOT EXISTS cust_device_link_3 (
  id BIGSERIAL PRIMARY KEY NOT NULL,
  cust_ix_link_id BIGINT NOT NULL REFERENCES cust_ix_link(id),
  device_id BIGINT NOT NULL REFERENCES device(id),
  first_seen TIMESTAMPTZ NOT NULL
);

CREATE UNIQUE INDEX IF NOT EXISTS idx_cust_device_link_3_cust_ix_link_id_device_id ON cust_device_link_3(cust_ix_link_id, device_id);


CREATE TABLE IF NOT EXISTS filepath (
  id BIGSERIAL PRIMARY KEY NOT NULL,
  path VARCHAR,
  sha256_path bytea NOT NULL CHECK(length(sha256_path) = 32),
  date_created TIMESTAMPTZ NOT NULL DEFAULT NOW()
);

CREATE UNIQUE INDEX IF NOT EXISTS idx_filepath_sha256_path ON filepath(sha256_path);

CREATE TABLE IF NOT EXISTS cust_filepath (
  id BIGSERIAL PRIMARY KEY NOT NULL,
  cust_ix_link_id BIGINT NOT NULL REFERENCES cust_ix_link(id),
  filepath_id BIGINT NOT NULL REFERENCES filepath(id),
  first_seen TIMESTAMPTZ NOT NULL,
  num_occurrences INT NOT NULL DEFAULT 1,
  date_modified TIMESTAMPTZ NOT NULL DEFAULT NOW()
);

CREATE UNIQUE INDEX IF NOT EXISTS idx_cust_filepath_cust_ix_link_id_filepath_id ON cust_filepath(cust_ix_link_id, filepath_id);

CREATE TABLE IF NOT EXISTS ubs_table_rotation (
    id BIGSERIAL PRIMARY KEY NOT NULL,
    table_name VARCHAR,
    previous VARCHAR,
    current VARCHAR,
    next VARCHAR
);

CREATE UNIQUE INDEX IF NOT EXISTS idx_ubs_table_rotation_table_name ON ubs_table_rotation(table_name);

INSERT INTO ubs_table_rotation(table_name, previous, current, next)
VALUES ('cust_device_link', 'cust_device_link_1', 'cust_device_link', 'cust_device_link_3')
ON CONFLICT (table_name) DO NOTHING ;

---------------------------------
-- SPs FOR INSERTION
---------------------------------

DO $$ BEGIN
    CREATE TYPE index_customer_metadata AS (sha256 bytea, md5 bytea, size BIGINT, org_key VARCHAR(128));
EXCEPTION
    WHEN duplicate_object THEN null;
END $$;

CREATE OR REPLACE FUNCTION insert_file_index_and_cust_link(insertion_data index_customer_metadata[])
  RETURNS VOID AS $$
  BEGIN
    INSERT INTO file_index (sha256, md5, size)
      (SELECT sha256, md5, size FROM unnest(insertion_data))
      ON CONFLICT (sha256) DO NOTHING;

    INSERT INTO cust(org_key, search_key)
        (SELECT org_key, digest(org_key, 'sha256') FROM unnest(insertion_data))
    ON CONFLICT (search_key) DO NOTHING;

    WITH data_cust(c_id, sha256) AS (
        SELECT c.id, i.sha256
        FROM unnest(insertion_data) i
        JOIN cust c ON c.search_key = digest(i.org_key, 'sha256')
    ), data_cust_cil(f_id, cust_id) AS (
        SELECT f.id, dc.c_id
        FROM data_cust dc
        JOIN file_index f ON f.sha256 = dc.sha256
    )
    INSERT INTO cust_ix_link(file_index_id, cust_id) (
        SELECT f_id, cust_id
        FROM data_cust_cil
    )
    ON CONFLICT (file_index_id, cust_id) DO NOTHING;
  END;
$$ LANGUAGE plpgsql;

DO $$ BEGIN
    CREATE TYPE static_metadata_insert_v3 AS (
      sha256 bytea,
      os_type VARCHAR,
      architecture SMALLINT,
      lang_id INT,
      charset_id INT,
      internal_name VARCHAR,
      product_name VARCHAR,
      company_name VARCHAR,
      copyright VARCHAR,
      trademark VARCHAR,
      file_description VARCHAR,
      file_version VARCHAR,
      comments VARCHAR,
      original_filename VARCHAR,
      product_description VARCHAR,
      product_version VARCHAR,
      private_build VARCHAR,
      special_build VARCHAR,
      sensor_version VARCHAR,
      internal_version SMALLINT,
      policy_version SMALLINT
    );
EXCEPTION
    WHEN duplicate_object THEN null;
END $$;

-- The APIs for tables company_name, program_name and copyright require that we return both NULL and '' (empty string) on the API.
-- xCBDEADBEEFBAFFEDDABBEDCBCBCBCBCBCBCB is the value that indicates NULL when we try to get a sha256 of a named value.
CREATE OR REPLACE FUNCTION insert_static_metadata_v2(insertion_data static_metadata_insert_v3[])
  RETURNS VOID AS $$
  BEGIN
    INSERT INTO program_name (name, sha256_name)
      (SELECT internal_name, (CASE WHEN (internal_name IS NOT NULL) THEN digest(internal_name, 'sha256') ELSE E'\\xCBDEADBEEFBAFFEDDABBEDCBCBCBCBCB' END) FROM unnest(insertion_data)
       UNION
       SELECT product_name, (CASE WHEN (product_name IS NOT NULL) THEN digest(product_name, 'sha256') ELSE E'\\xCBDEADBEEFBAFFEDDABBEDCBCBCBCBCB' END) FROM unnest(insertion_data)
       )
      ON CONFLICT (sha256_name) DO NOTHING;

    INSERT INTO company_name (name, sha256_name)
      SELECT company_name, (CASE WHEN (company_name IS NOT NULL) THEN digest(company_name, 'sha256') ELSE E'\\xCBDEADBEEFBAFFEDDABBEDCBCBCBCBCB' END) FROM unnest(insertion_data)
      ON CONFLICT (sha256_name) DO NOTHING;

    INSERT INTO copyright (name, sha256_name)
      (SELECT copyright, (CASE WHEN (copyright IS NOT NULL) THEN digest(copyright, 'sha256') ELSE E'\\xCBDEADBEEFBAFFEDDABBEDCBCBCBCBCB' END) FROM unnest(insertion_data)
       UNION
       SELECT trademark, (CASE WHEN (trademark IS NOT NULL) THEN digest(trademark, 'sha256') ELSE E'\\xCBDEADBEEFBAFFEDDABBEDCBCBCBCBCB' END) FROM unnest(insertion_data)
       )
      ON CONFLICT (sha256_name) DO NOTHING;

    INSERT INTO metadata (
        file_index_id,
        internal_name_id,
        product_name_id,
        company_name_id,
        copyright_id,
        trademark_id,
        file_description,
        file_version,
        comments,
        original_filename,
        product_description,
        product_version,
        private_build,
        special_build,
        os_type,
        architecture,
        lang_id,
        charset_id,
        sensor_version,
        internal_version,
        policy_version)
      (SELECT fi.id, pn1.id, pn2.id, cn.id, c1.id, c2.id, i.file_description, i.file_version, i.comments,
              i.original_filename, i.product_description, i.product_version, i.private_build, i.special_build,
              i.os_type, i.architecture, i.lang_id, i.charset_id, i.sensor_version, i.internal_version, i.policy_version
        FROM unnest(insertion_data) i
        JOIN file_index fi ON i.sha256 = fi.sha256
        JOIN program_name pn1 ON (CASE WHEN (i.internal_name IS NOT NULL) THEN digest(i.internal_name, 'sha256') ELSE E'\\xCBDEADBEEFBAFFEDDABBEDCBCBCBCBCB' END) = pn1.sha256_name
        JOIN program_name pn2 ON (CASE WHEN (i.product_name IS NOT NULL) THEN digest(i.product_name, 'sha256') ELSE E'\\xCBDEADBEEFBAFFEDDABBEDCBCBCBCBCB' END) = pn2.sha256_name
        JOIN company_name cn ON (CASE WHEN (i.company_name IS NOT NULL) THEN digest(i.company_name, 'sha256') ELSE E'\\xCBDEADBEEFBAFFEDDABBEDCBCBCBCBCB' END) = cn.sha256_name
        JOIN copyright c1 ON (CASE WHEN (i.copyright IS NOT NULL) THEN digest(i.copyright, 'sha256') ELSE E'\\xCBDEADBEEFBAFFEDDABBEDCBCBCBCBCB' END) = c1.sha256_name
        JOIN copyright c2 ON (CASE WHEN (i.trademark IS NOT NULL) THEN digest(i.trademark, 'sha256') ELSE E'\\xCBDEADBEEFBAFFEDDABBEDCBCBCBCBCB' END) = c2.sha256_name
      )
      ON CONFLICT (file_index_id) DO UPDATE
        SET internal_name_id = EXCLUDED.internal_name_id,
            product_name_id = EXCLUDED.product_name_id,
            company_name_id = EXCLUDED.company_name_id,
            copyright_id = EXCLUDED.copyright_id,
            trademark_id = EXCLUDED.trademark_id,
            file_description = EXCLUDED.file_description,
            file_version = EXCLUDED.file_version,
            comments = EXCLUDED.comments,
            original_filename = EXCLUDED.original_filename,
            product_description = EXCLUDED.product_description,
            product_version = EXCLUDED.product_version,
            private_build = EXCLUDED.private_build,
            special_build = EXCLUDED.special_build,
            os_type = EXCLUDED.os_type,
            architecture = EXCLUDED.architecture,
            lang_id = EXCLUDED.lang_id,
            charset_id = EXCLUDED.charset_id,
            sensor_version = EXCLUDED.sensor_version,
            internal_version = EXCLUDED.internal_version,
            policy_version = EXCLUDED.policy_version;
  END;
$$ LANGUAGE plpgsql;

DO $$ BEGIN
    CREATE TYPE filepath_insert_v2 AS (sha256 bytea, org_key VARCHAR(128), filepath VARCHAR, num_occurrences SMALLINT,
                                first_seen TIMESTAMPTZ);
EXCEPTION
    WHEN duplicate_object THEN null;
END $$;

CREATE OR REPLACE FUNCTION insert_filepath_v2(insertion_data filepath_insert_v2[])
  RETURNS VOID AS $$
  BEGIN
    INSERT INTO filepath (path, sha256_path)
      (SELECT filepath, (CASE WHEN (filepath IS NOT NULL) THEN digest(filepath, 'sha256') ELSE E'\\xCBDEADBEEFBAFFEDDABBEDCBCBCBCBCBCBDEADBEEFBAFFEDDABBEDCBCBCBCBCB' END)
       FROM unnest(insertion_data))
      ON CONFLICT (sha256_path) DO NOTHING;

    WITH data1(c_id, fi_id, f_id, num_occurrences, first_seen) AS (
      SELECT c.id, fi.id, f.id, i.num_occurrences, i.first_seen
      FROM unnest(insertion_data) i
      JOIN filepath f ON (CASE WHEN (i.filepath IS NOT NULL) THEN digest(i.filepath, 'sha256') ELSE E'\\xCBDEADBEEFBAFFEDDABBEDCBCBCBCBCBCBDEADBEEFBAFFEDDABBEDCBCBCBCBCB' END) = f.sha256_path
      JOIN cust c ON digest(i.org_key, 'sha256') = c.search_key
      JOIN file_index fi ON i.sha256 = fi.sha256
    ), data2(cixl_id, f_id, num_occurrences, first_seen) AS (
      SELECT cixl.id, d.f_id, d.num_occurrences, d.first_seen
      FROM data1 d
      JOIN cust_ix_link cixl ON d.c_id = cixl.cust_id AND d.fi_id = cixl.file_index_id
    )
      INSERT INTO cust_filepath (cust_ix_link_id, filepath_id, num_occurrences, first_seen)
      (
        SELECT d2.cixl_id, d2.f_id, d2.num_occurrences, d2.first_seen
        FROM data2 d2
      )
      ON CONFLICT (cust_ix_link_id, filepath_id) DO UPDATE
        SET num_occurrences = cust_filepath.num_occurrences + EXCLUDED.num_occurrences,
        date_modified = NOW();

  END;
$$ LANGUAGE plpgsql;

DO $$ BEGIN
    CREATE TYPE device_insert_v2 AS (sha256 bytea, org_key VARCHAR(128), device_id BIGINT, device_name VARCHAR(256),
                              first_seen TIMESTAMPTZ);
EXCEPTION
    WHEN duplicate_object THEN null;
END $$;

CREATE OR REPLACE FUNCTION insert_device_v2(insertion_data device_insert_v2[])
  RETURNS bytea[] AS $$
  DECLARE ret bytea[];
  BEGIN
    INSERT INTO device (device_id, device_name)
      (SELECT device_id, device_name FROM unnest(insertion_data))
      ON CONFLICT (device_id, device_name) DO NOTHING;

    WITH data1(c_id, fi_id, d_id, sha256, first_seen) AS (
    SELECT c.id, fi.id, d.id, i.sha256, i.first_seen
       FROM unnest(insertion_data) i
       JOIN device d ON i.device_id = d.device_id AND i.device_name = d.device_name
       JOIN cust c ON digest(i.org_key, 'sha256') = c.search_key
       JOIN file_index fi ON i.sha256 = fi.sha256
    ), data2(cixl_id, d_id, sha256, first_seen) AS (
    SELECT cixl.id, d_id, sha256, first_seen
        FROM data1 d
        JOIN cust_ix_link cixl ON c_id = cixl.cust_id AND d.fi_id = cixl.file_index_id
    ), ins (cixl_id, d_id) AS (
      INSERT INTO cust_device_link(cust_ix_link_id, device_id, first_seen)
      SELECT cixl_id, d_id, first_seen
      FROM data2
      ON CONFLICT (cust_ix_link_id, device_id) DO NOTHING
      RETURNING cust_ix_link_id, device_id
    )
    SELECT array(
      SELECT data2.sha256
      FROM data2
      EXCEPT (
        SELECT data2.sha256
        FROM data2
        JOIN ins ON data2.cixl_id = ins.cixl_id AND data2.d_id = ins.d_id
      )
    )
    INTO ret;
    RETURN ret;

  END;
$$ LANGUAGE plpgsql;

DO $$ BEGIN
    CREATE TYPE device_insert_v3 AS (sha256 bytea, org_key VARCHAR(128), device_id BIGINT, device_name VARCHAR(256),
                              first_seen TIMESTAMPTZ);
EXCEPTION
    WHEN duplicate_object THEN null;
END $$;

CREATE OR REPLACE FUNCTION insert_device_v3(insertion_data device_insert_v3[])
  RETURNS bytea[] AS $$
  DECLARE ret bytea[];
  BEGIN
    INSERT INTO device (device_id, device_name)
      (SELECT device_id, device_name FROM unnest(insertion_data))
      ON CONFLICT (device_id, device_name) DO UPDATE SET last_seen = NOW();

    WITH data1(c_id, fi_id, d_id, sha256, first_seen) AS (
    SELECT c.id, fi.id, d.id, i.sha256, i.first_seen
       FROM unnest(insertion_data) i
       JOIN device d ON i.device_id = d.device_id AND i.device_name = d.device_name
       JOIN cust c ON digest(i.org_key, 'sha256') = c.search_key
       JOIN file_index fi ON i.sha256 = fi.sha256
    ), data2(cixl_id, d_id, sha256, first_seen) AS (
    SELECT cixl.id, d_id, sha256, first_seen
        FROM data1 d
        JOIN cust_ix_link cixl ON c_id = cixl.cust_id AND d.fi_id = cixl.file_index_id
    ), ins (cixl_id, d_id) AS (
      INSERT INTO cust_device_link(cust_ix_link_id, device_id, first_seen)
      SELECT cixl_id, d_id, first_seen
      FROM data2
      ON CONFLICT (cust_ix_link_id, device_id) DO NOTHING
      RETURNING cust_ix_link_id, device_id
    )
    SELECT array(
      SELECT data2.sha256
      FROM data2
      EXCEPT (
        SELECT data2.sha256
        FROM data2
        JOIN ins ON data2.cixl_id = ins.cixl_id AND data2.d_id = ins.d_id
      )
    )
    INTO ret;
    RETURN ret;

  END;
$$ LANGUAGE plpgsql;


-- write to device and cust_device_link_1
CREATE OR REPLACE FUNCTION insert_device_1_v3(insertion_data device_insert_v3[])
  RETURNS bytea[] AS $$
  DECLARE ret bytea[];
  BEGIN
    INSERT INTO device (device_id, device_name)
      (SELECT device_id, device_name FROM unnest(insertion_data))
      ON CONFLICT (device_id, device_name) DO UPDATE SET last_seen = NOW();

    WITH data1(c_id, fi_id, d_id, sha256, first_seen) AS (
    SELECT c.id, fi.id, d.id, i.sha256, i.first_seen
       FROM unnest(insertion_data) i
       JOIN device d ON i.device_id = d.device_id AND i.device_name = d.device_name
       JOIN cust c ON digest(i.org_key, 'sha256') = c.search_key
       JOIN file_index fi ON i.sha256 = fi.sha256
    ), data2(cixl_id, d_id, sha256, first_seen) AS (
    SELECT cixl.id, d_id, sha256, first_seen
        FROM data1 d
        JOIN cust_ix_link cixl ON c_id = cixl.cust_id AND d.fi_id = cixl.file_index_id
    ), ins (cixl_id, d_id) AS (
      INSERT INTO cust_device_link_1(cust_ix_link_id, device_id, first_seen)
      SELECT cixl_id, d_id, first_seen
      FROM data2
      ON CONFLICT (cust_ix_link_id, device_id) DO NOTHING
      RETURNING cust_ix_link_id, device_id
    )
    SELECT array(
      SELECT data2.sha256
      FROM data2
      EXCEPT (
        SELECT data2.sha256
        FROM data2
        JOIN ins ON data2.cixl_id = ins.cixl_id AND data2.d_id = ins.d_id
      )
    )
    INTO ret;
    RETURN ret;

  END;
$$ LANGUAGE plpgsql;


-- write to device and cust_device_link_3 as rotation/cache table
CREATE OR REPLACE FUNCTION insert_device_3_v3(insertion_data device_insert_v3[])
  RETURNS bytea[] AS $$
  DECLARE ret bytea[];
  BEGIN
    INSERT INTO device (device_id, device_name)
      (SELECT device_id, device_name FROM unnest(insertion_data))
      ON CONFLICT (device_id, device_name) DO UPDATE SET last_seen = NOW();

    WITH data1(c_id, fi_id, d_id, sha256, first_seen) AS (
    SELECT c.id, fi.id, d.id, i.sha256, i.first_seen
       FROM unnest(insertion_data) i
       JOIN device d ON i.device_id = d.device_id AND i.device_name = d.device_name
       JOIN cust c ON digest(i.org_key, 'sha256') = c.search_key
       JOIN file_index fi ON i.sha256 = fi.sha256
    ), data2(cixl_id, d_id, sha256, first_seen) AS (
    SELECT cixl.id, d_id, sha256, first_seen
        FROM data1 d
        JOIN cust_ix_link cixl ON c_id = cixl.cust_id AND d.fi_id = cixl.file_index_id
    ), ins (cixl_id, d_id) AS (
      INSERT INTO cust_device_link_3(cust_ix_link_id, device_id, first_seen)
      SELECT cixl_id, d_id, first_seen
      FROM data2
      ON CONFLICT (cust_ix_link_id, device_id) DO NOTHING
      RETURNING cust_ix_link_id, device_id
    )
    SELECT array(
      SELECT data2.sha256
      FROM data2
      EXCEPT (
        SELECT data2.sha256
        FROM data2
        JOIN ins ON data2.cixl_id = ins.cixl_id AND data2.d_id = ins.d_id
      )
    )
    INTO ret;
    RETURN ret;

  END;
$$ LANGUAGE plpgsql;


DO $$ BEGIN
    CREATE TYPE signature_insert_v4 AS (
      sha256 bytea,
      org_key VARCHAR(128),
      publisher_name VARCHAR,
      issuer_name VARCHAR,
      first_seen TIMESTAMPTZ,
      sign_timestamp TIMESTAMPTZ,
      sign_state SMALLINT,
      num_occurrences INT
    );
EXCEPTION
    WHEN duplicate_object THEN null;
END $$;

CREATE OR REPLACE FUNCTION insert_signature_v4(insertion_data signature_insert_v4[])
  RETURNS VOID AS $$
BEGIN
   -- Add the publisher and issuer names
    INSERT INTO company_name (name, sha256_name)
    (
      SELECT publisher_name, (CASE WHEN (publisher_name IS NOT NULL) THEN digest(publisher_name, 'sha256') ELSE E'\\xCBDEADBEEFBAFFEDDABBEDCBCBCBCBCB' END) FROM unnest(insertion_data)
      UNION SELECT issuer_name, (CASE WHEN (issuer_name IS NOT NULL) THEN digest(issuer_name, 'sha256') ELSE E'\\xCBDEADBEEFBAFFEDDABBEDCBCBCBCBCB' END) FROM unnest(insertion_data)
    ) ON CONFLICT (sha256_name) DO NOTHING;

    -- Add the signature data
    INSERT INTO signature (
      digsig_publisher_name_id,
      digsig_issuer_name_id, digsig_sign_time)
    (
      SELECT cn.id,
        cn2.id, i.sign_timestamp
      FROM unnest(insertion_data) i
      JOIN company_name cn ON (CASE WHEN (i.publisher_name IS NOT NULL) THEN digest(i.publisher_name, 'sha256') ELSE E'\\xCBDEADBEEFBAFFEDDABBEDCBCBCBCBCB' END) = cn.sha256_name
      JOIN company_name cn2 ON (CASE WHEN (i.issuer_name IS NOT NULL) THEN digest(i.issuer_name, 'sha256') ELSE E'\\xCBDEADBEEFBAFFEDDABBEDCBCBCBCBCB' END) = cn2.sha256_name
    ) ON CONFLICT DO NOTHING;

    -- Add data to the link table
    WITH f_cn_cn2_cust (c_id, fi_id, sign_id, sign_state, is_catalog_signature, first_seen, num_occurrences) AS (
        -- Work with the !NULL
        SELECT c.id, fi.id, s.id, i.sign_state, ((i.sign_state & 128) = 128), i.first_seen, i.num_occurrences
        FROM unnest(insertion_data) i
        JOIN file_index fi ON i.sha256 = fi.sha256
        JOIN company_name cn ON (CASE WHEN (i.publisher_name IS NOT NULL) THEN digest(i.publisher_name, 'sha256') ELSE E'\\xCBDEADBEEFBAFFEDDABBEDCBCBCBCBCB' END) = cn.sha256_name
        JOIN company_name cn2 ON (CASE WHEN (i.issuer_name IS NOT NULL) THEN digest(i.issuer_name, 'sha256') ELSE E'\\xCBDEADBEEFBAFFEDDABBEDCBCBCBCBCB' END) = cn2.sha256_name
        JOIN cust c ON digest(i.org_key, 'sha256') = c.search_key
        JOIN signature s ON s.digsig_publisher_name_id = cn.id
                        AND s.digsig_issuer_name_id = cn2.id
                        AND s.digsig_sign_time = i.sign_timestamp
        WHERE i.sign_timestamp NOTNULL

        UNION ALL

        -- Work with the NULL
        SELECT c.id, fi.id, s.id, i.sign_state, ((i.sign_state & 128) = 128), i.first_seen, i.num_occurrences
        FROM unnest(insertion_data) i
        JOIN file_index fi ON i.sha256 = fi.sha256
        JOIN company_name cn ON (CASE WHEN (i.publisher_name IS NOT NULL) THEN digest(i.publisher_name, 'sha256') ELSE E'\\xCBDEADBEEFBAFFEDDABBEDCBCBCBCBCB' END) = cn.sha256_name
        JOIN company_name cn2 ON (CASE WHEN (i.issuer_name IS NOT NULL) THEN digest(i.issuer_name, 'sha256') ELSE E'\\xCBDEADBEEFBAFFEDDABBEDCBCBCBCBCB' END) = cn2.sha256_name
        JOIN cust c ON digest(i.org_key, 'sha256') = c.search_key
        JOIN signature s ON s.digsig_publisher_name_id = cn.id
                        AND s.digsig_issuer_name_id = cn2.id
                        AND (s.digsig_sign_time ISNULL and i.sign_timestamp ISNULL)
        WHERE i.sign_timestamp ISNULL
    ), f_cn_cn2_cust_cil (cil_id, sig_id, sig_state, is_catalog_signature, first_seen, num_occurrences) AS (
        SELECT cil.id, tmp1.sign_id, tmp1.sign_state, tmp1.is_catalog_signature, tmp1.first_seen, tmp1.num_occurrences
        FROM f_cn_cn2_cust tmp1
        JOIN cust_ix_link cil ON tmp1.c_id = cil.cust_id AND tmp1.fi_id = cil.file_index_id
    )
    INSERT INTO signature_result_link (
      cust_ix_link_id, signature_id, sig_state, is_catalog_signature, first_seen, num_occurrences
    )(
        SELECT cil_id, sig_id, sig_state, is_catalog_signature, first_seen, num_occurrences
        FROM f_cn_cn2_cust_cil
    ) ON CONFLICT (cust_ix_link_id, signature_id, sig_state)
        DO UPDATE
        SET num_occurrences = signature_result_link.num_occurrences + EXCLUDED.num_occurrences,
          date_modified = NOW();
  END;
$$ LANGUAGE plpgsql;

DO $$ BEGIN
    CREATE TYPE file_index_cust_partner AS (
      sha256 bytea,
      org_key VARCHAR,
      partner SMALLINT
    );
EXCEPTION
    WHEN duplicate_object THEN null;
END $$;

CREATE OR REPLACE FUNCTION insert_file_index_cust_partner_link(insertion_data file_index_cust_partner[])
  RETURNS VOID AS $$
BEGIN
  INSERT INTO file_index_cust_partner_link (file_index_id, cust_id, partner_id)
    (SELECT fi.id, c.id, i.partner
     FROM unnest(insertion_data) i
     JOIN file_index fi ON i.sha256 = fi.sha256
     JOIN cust c ON digest(i.org_key, 'sha256') = c.search_key)
  ON CONFLICT (file_index_id, partner_id)
  DO NOTHING;
END;
  $$ LANGUAGE plpgsql;

DO $$ BEGIN
    CREATE TYPE hash_cust_device_v3 AS (
        sha256      VARCHAR,
        org_key     VARCHAR,
        device_id   BIGINT,
        device_name VARCHAR
    );
EXCEPTION
    WHEN duplicate_object THEN null;
END $$;

CREATE OR REPLACE FUNCTION select_duplicate_cust_device_links_v3(insertion_data hash_cust_device_v3[])
    RETURNS TABLE (sha256 varchar, org_key varchar, device_id bigint, device_name varchar) AS $$
BEGIN
    RETURN QUERY
    WITH data1(c_id, fi_id, d_id, sha256, org_key, device_id, device_name) AS (
      SELECT c.id, fi.id, d.id, i.sha256, i.org_key, i.device_id, i.device_name
      FROM unnest(insertion_data) i
      JOIN device d ON i.device_id = d.device_id AND i.device_name = d.device_name
      JOIN cust c ON digest(i.org_key, 'sha256') = c.search_key
      JOIN file_index fi ON decode(i.sha256, 'hex') = fi.sha256
    ), data2(cixl_id, d_id, sha256, org_key, device_id, device_name) AS (
      SELECT cixl.id, d.d_id, d.sha256, d.org_key, d.device_id, d.device_name
      FROM data1 d
      JOIN cust_ix_link cixl ON d.c_id = cixl.cust_id AND d.fi_id = cixl.file_index_id
    )
    SELECT DISTINCT
      d.sha256
      , d.org_key
      , d.device_id
      , d.device_name
    FROM data2 d
    JOIN cust_device_link cdl ON d.cixl_id = cdl.cust_ix_link_id AND d.d_id = cdl.device_id;
END;
$$ LANGUAGE plpgsql;

-- check cust_device_link_1, trade redundency for performance
CREATE OR REPLACE FUNCTION select_duplicate_cust_device_links_1_v3(insertion_data hash_cust_device_v3[])
    RETURNS TABLE (sha256 varchar, org_key varchar, device_id bigint, device_name varchar) AS $$
BEGIN
    RETURN QUERY
    WITH data1(c_id, fi_id, d_id, sha256, org_key, device_id, device_name) AS (
      SELECT c.id, fi.id, d.id, i.sha256, i.org_key, i.device_id, i.device_name
      FROM unnest(insertion_data) i
      JOIN device d ON i.device_id = d.device_id AND i.device_name = d.device_name
      JOIN cust c ON digest(i.org_key, 'sha256') = c.search_key
      JOIN file_index fi ON decode(i.sha256, 'hex') = fi.sha256
    ), data2(cixl_id, d_id, sha256, org_key, device_id, device_name) AS (
      SELECT cixl.id, d.d_id, d.sha256, d.org_key, d.device_id, d.device_name
      FROM data1 d
      JOIN cust_ix_link cixl ON d.c_id = cixl.cust_id AND d.fi_id = cixl.file_index_id
    )
    SELECT DISTINCT
      d.sha256
      , d.org_key
      , d.device_id
      , d.device_name
    FROM data2 d
    JOIN cust_device_link_1 cdl ON d.cixl_id = cdl.cust_ix_link_id AND d.d_id = cdl.device_id;
END;
$$ LANGUAGE plpgsql;

-- check cust_device_link_3
CREATE OR REPLACE FUNCTION select_duplicate_cust_device_links_3_v3(insertion_data hash_cust_device_v3[])
    RETURNS TABLE (sha256 varchar, org_key varchar, device_id bigint, device_name varchar) AS $$
BEGIN
    RETURN QUERY
    WITH data1(c_id, fi_id, d_id, sha256, org_key, device_id, device_name) AS (
      SELECT c.id, fi.id, d.id, i.sha256, i.org_key, i.device_id, i.device_name
      FROM unnest(insertion_data) i
      JOIN device d ON i.device_id = d.device_id AND i.device_name = d.device_name
      JOIN cust c ON digest(i.org_key, 'sha256') = c.search_key
      JOIN file_index fi ON decode(i.sha256, 'hex') = fi.sha256
    ), data2(cixl_id, d_id, sha256, org_key, device_id, device_name) AS (
      SELECT cixl.id, d.d_id, d.sha256, d.org_key, d.device_id, d.device_name
      FROM data1 d
      JOIN cust_ix_link cixl ON d.c_id = cixl.cust_id AND d.fi_id = cixl.file_index_id
    )
    SELECT DISTINCT
      d.sha256
      , d.org_key
      , d.device_id
      , d.device_name
    FROM data2 d
    JOIN cust_device_link_3 cdl ON d.cixl_id = cdl.cust_ix_link_id AND d.d_id = cdl.device_id;
END;
$$ LANGUAGE plpgsql;